/*
  Calisto - An internet Talker
  Copyright (C) 1998-2001 Peter Howkins, Matthew Howkins and Simon Howkins

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,     
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License   
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Email: calisto@calisto.cx

  $Id: inifile2.c,v 1.3 2002/12/02 19:47:11 peter Exp $

 */
static const char rcsid[] = "$Id: inifile2.c,v 1.3 2002/12/02 19:47:11 peter Exp $";

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "config.h"
#include "debug.h"
#include "inifile2.h"
#include "mem.h"
#include "msnprintf.h"
#include "strplus.h"

static void inifile_free(INIFILE *ini)
{
  inifile_line *line, *temp;

  DBUGF("> inifile_free()\n");
  ASSERT(ini);

DEBUG_ONLY(
  inifile_test(ini);
);

  /* step through linked list freeing lines */
  line = ini->next;
  while(line) {
    temp = line;
    line = line->next;
    MEM_FREE(temp);
  }
  /* free inifile_hdr */
  MEM_FREE(ini);
  DBUGF("<\n");
}

INIFILE *inifile_open(const char *filename)
{
  unsigned lineno = 0;
  char line[1024];
  FILE *temp;
  INIFILE *ini;
  inifile_line **blah = NULL;

  DBUGF("> inifile_open()\n");

  temp = fopen(filename, "r");
  if(!temp) {
    /* file does not exist attempt to make one */
    temp = fopen(filename, "w+");
    if(!temp) {
      DBUGF("< failed to open file\n");
      return NULL;
    }
  }
  
  /* set up the inifile header */
  ini = MEM_MALLOC(sizeof(struct inifile_hdr));
  if(!ini) {
    fprintf(stderr, "not enough memory (inifile_hdr)\n");
    DBUGF("< out of memory\n");
    return NULL;
  }
  STRNCOPY(ini->path, filename, sizeof(ini->path));
  ini->next = NULL;

  /* record the pointer to the next line, so we can copy the address of the
     first line to it, once we have read it in */
  blah = &ini->next;

  rewind(temp);
  while (fgets(line, sizeof(line), temp)) {
    char *token;
    char *gap;
    char *val;
    inifile_line *iniline;

    lineno++;

    /* All but the last line will probably have a newline on the end */
    strchomp(line);

    /* Remove leading whitespace */
    token = line;
    while (*token && isspace(*token))
      token++;

    /* Ignore comments and blank lines */
    if (*token == '\0' || *token == '#')
      continue;

    /* Find end of token */
    gap = token;
    while (*gap && (isalpha(*gap) || *gap == '_'))
      gap++;
    val = gap;

    /* Ignore whitespace (before =) */
    while (*val && isspace(*val))
      val++;

    /* Check for = */
    if (*val++ != '=') {
      fprintf(stderr, "inifile: error in '%s' line %u\n", filename, lineno);
      continue;
    }

    /* Ignore more whitespace (after =) */
    while (*val && isspace(*val))
      val++;
    
    /* Terminate token with '\0' */
    *gap = '\0';


    /* make an inifile line */
    iniline = MEM_MALLOC(sizeof(inifile_line));
    if(!iniline) {
      /* destroy the other lines and the inifile_hdr */
      inifile_free(ini);

      fprintf(stderr, "not enough memory (inifile_line)\n");
      DBUGF("< out of memory 2\n");
      return NULL;
    }
    STRNCOPY(iniline->token, token, TOKEN_LEN);
    STRNCOPY(iniline->val, val, VAL_LEN);
    iniline->next = NULL;
 
    *blah = iniline;

    blah = &iniline->next; 
  }
  fclose(temp);

  DBUGF("<\n");
  return ini;
}

void inifile_test(INIFILE *ini)
{
  inifile_line *line;

  if(!ini) {
    fprintf(stderr, "null inifile passed to inifile_test\n");
    return;
  }
  line = ini->next;

  while(line) {
    printf("%s %s\n", line->token, line->val);

    line = line->next;
  }
}

void inifile_close(INIFILE *ini, int mode)
{
  inifile_line *line;
  FILE *tempfile;

  ASSERT(ini);

  if(mode == INIFILE_DO_WRITE) {
    /* Open file, step though linked list, write out line */
    tempfile = fopen(ini->path, "w");
    if(!tempfile) {
      /* memory cleanup */
      inifile_free(ini);

      fprintf(stderr, "Error opening '%s'\n", ini->path);
      return;
    }

    line = ini->next;
    while(line) {
      /* output line */   
      fprintf(tempfile, "%s = %s\n", line->token, line->val);

      line = line->next;
    }
    fclose(tempfile);
  }

  /* free memory */
  inifile_free(ini);
}

int inifile_get_token(INIFILE *ini, const char *token_name,
                      const char *format, void *value, 
                      const void *default_value)
{
  inifile_line *line;

  /* does the ini file exist ? */
  if (!ini)
    return 1;

  /* check for token in list and fill in details */
  line = ini->next;
  while(line) {
    /* output line */
    if (STRIEQ(line->token, token_name)) {
      if (STREQ(format, "%s")) {
        strcpy(value, line->val);  
      } else {
        sscanf(line->val, format, value);
      }
      return 0;
    }
    line = line->next;
  }

  /* token not found return default */
  if (STREQ(format, "%s")) {
    strcpy(value, default_value);
  } else if (STREQ(format, "%d") || STREQ(format, "%i")) {
    *((int *) value) = *((int *) default_value);
  } else if (STREQ(format, "%u") || STREQ(format, "%x")) {
    *((unsigned *) value) = *((unsigned *) default_value);
  } else if (STREQ(format, "%ld") || STREQ(format, "%li")) {
    *((long *) value) = *((long *) default_value);
  } else if (STREQ(format, "%lu") || STREQ(format, "%lx")) {
    *((unsigned long *) value) = *((unsigned long *) default_value);
  } else if (STREQ(format, "%f") || STREQ(format, "%e") || STREQ(format, "%g")) {
    *((double *) value) = *((double *) default_value);
  } else if (STREQ(format, "%lf") || STREQ(format, "%le") || STREQ(format, "%lg")) {
    *((long double *) value) = *((long double *) default_value);
  } else if (STREQ(format, "%c")) {
    *((char *) value) = *((char *) default_value);
  } else {
    fprintf(stderr, "type '%s' not handled by inifile system\n", format); 
    return 1;
  }
  return 0;
}

/* set token
  if inifile ptr is null, return error;
  
search through list to find token_name, list may be no elements if 
new inifile

if(found) {
  change val using sprintf(val, format, value)
} else {
  malloc a line, and insert on end of list
}

int inifile_set_token(const char *token_name, const char *format, void *value)
*/
int inifile_set_token(INIFILE *ini, const char *token_name,
                      const char *format, void *value)
{
  inifile_line *line, *iniline;
  inifile_line **blah = NULL;

  /* does the ini file exist ? */
  if (!ini)
    return 1;

  /* check for token in list and fill in details */
  line = ini->next;
  blah = &ini->next;
  while(line) {
    /* output line */
    if (STRIEQ(line->token, token_name)) {
      if (STREQ(format, "%s")) {
        msnprintf(line->val, VAL_LEN, format, (char *) value);
      } else if (STREQ(format, "%d") || STREQ(format, "%i")) {
        msnprintf(line->val, VAL_LEN, format, *((int *) value));
      } else if (STREQ(format, "%u") || STREQ(format, "%x")) {
        msnprintf(line->val, VAL_LEN, format, *((unsigned *) value));
      } else if (STREQ(format, "%ld") || STREQ(format, "%li")) {
        msnprintf(line->val, VAL_LEN, format, *((long *) value));
      } else if (STREQ(format, "%lu") || STREQ(format, "%lx")) {
        msnprintf(line->val, VAL_LEN, format, *((unsigned long *) value));
      } else if (STREQ(format, "%f") || STREQ(format, "%e") || STREQ(format, "%g")) {
        msnprintf(line->val, VAL_LEN, format, *((double *) value));
      } else if (STREQ(format, "%lf") || STREQ(format, "%le") || STREQ(format, "%lg")) {
        msnprintf(line->val, VAL_LEN, format, *((long double *) value));
      } else if (STREQ(format, "%c")) {
        msnprintf(line->val, VAL_LEN, format, *((char *) value));
      } else {
        fprintf(stderr, "type '%s' not handled by inifile system\n", format); 
        return 1;
      }
      return 0;
    }
    blah = &line->next;
    line = line->next;
  }

  /* make another line on the end of the list and fill in the details */
  iniline = MEM_MALLOC(sizeof(inifile_line));
  if(!iniline) {
    /* do not destruct the rest of the inifile struct as the rest should still
       be useable */
    fprintf(stderr, "not enough memory (inifile_line)\n");
    return 1;
  }

  STRNCOPY(iniline->token, token_name, TOKEN_LEN);
  if (STREQ(format, "%s")) {
    msnprintf(iniline->val, VAL_LEN, format, (char *) value);
  } else if (STREQ(format, "%d") || STREQ(format, "%i")) {
    msnprintf(iniline->val, VAL_LEN, format, *((int *) value));
  } else if (STREQ(format, "%u") || STREQ(format, "%x")) {
    msnprintf(iniline->val, VAL_LEN, format, *((unsigned *) value));
  } else if (STREQ(format, "%ld") || STREQ(format, "%li")) {
    msnprintf(iniline->val, VAL_LEN, format, *((long *) value));
  } else if (STREQ(format, "%lu") || STREQ(format, "%lx")) {
    msnprintf(iniline->val, VAL_LEN, format, *((unsigned long *) value));
  } else if (STREQ(format, "%f") || STREQ(format, "%e") || STREQ(format, "%g")) {
    msnprintf(iniline->val, VAL_LEN, format, *((double *) value));
  } else if (STREQ(format, "%lf") || STREQ(format, "%le") || STREQ(format, "%lg")) {
    msnprintf(iniline->val, VAL_LEN, format, *((long double *) value));
  } else if (STREQ(format, "%c")) {
    msnprintf(iniline->val, VAL_LEN, format, *((char *) value));
  } else {
    fprintf(stderr, "type '%s' not handled by inifile system\n", format); 
    return 1;
  }

  iniline->next = NULL;

  /* link in the node */
  *blah = iniline;

  return 0;
}

